/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package guru.zoroark.tegral.openapi.scripthost

import com.google.common.jimfs.Configuration
import com.google.common.jimfs.Jimfs
import io.swagger.v3.oas.models.OpenAPI
import io.swagger.v3.oas.models.info.Info
import kotlinx.coroutines.runBlocking
import java.io.File
import java.nio.file.Files
import kotlin.script.experimental.api.ResultWithDiagnostics
import kotlin.test.Test
import kotlin.test.assertEquals
import kotlin.test.assertTrue

class OpenApiScriptHostTest {
    @Test
    fun `Compile and evaluate script from path`() {
        val fs = Jimfs.newFileSystem(Configuration.unix())
        val path = fs.getPath("/example.openapi.kts")
        Files.writeString(
            path,
            """
            title = "My API"
            version = "0.0.0"
            description = "Hello there!"
            """.trimIndent()
        )

        val messages = mutableListOf<String>()
        val openApi = runBlocking { OpenApiScriptHost.compileScript(path) { messages += it } }
        assertTrue(openApi is ResultWithDiagnostics.Success)
        assertEquals(
            OpenAPI().apply {
                info = Info().apply {
                    title = "My API"
                    version = "0.0.0"
                    description = "Hello there!"
                }
            },
            openApi.value
        )
    }

    @Test
    fun `Compile and evaluate script from file`() {
        val file = File.createTempFile("tegral-openapi-scripthost-test", ".openapi.kts")
        file.writeText(
            """
            title = "My API"
            version = "0.0.0"
            description = "Hello there!"
            """.trimIndent()
        )
        file.deleteOnExit()

        val messages = mutableListOf<String>()
        val openApi = runBlocking { OpenApiScriptHost.compileScript(file) { messages += it } }
        assertTrue(openApi is ResultWithDiagnostics.Success)
        assertEquals(
            OpenAPI().apply {
                info = Info().apply {
                    title = "My API"
                    version = "0.0.0"
                    description = "Hello there!"
                }
            },
            openApi.value
        )
    }
}
